/* This code is adapted from http://www.dogma.net/markn/articles/pq_stl/priority.htm */

#include <iostream>
#include <iomanip>
#include <fstream>
#include <vector>
#include <stack>
#include <string>
#include <queue>
using namespace std;


// To append codes

unsigned SetBit(unsigned x, unsigned pos, unsigned bit) {
      return x | (bit << pos);
}

// codetable := new TCodeEntry[256]

class TCodeEntry {
public:
   unsigned count;
   unsigned bits;
   unsigned code;
   TCodeEntry() {count=0;bits=0;code=0u;};
};

//
// The node class is used to represent both leaf
// and internal nodes. leaf nodes have 0s in the
// child pointers, and their value member corresponds
// to the character they encode.  internal nodes
// don't have anything meaningful in their value
// member, but their child pointers point to
// other nodes.
//

class node {
private:
    int weight;
    unsigned char value;
    const node *child0;
    const node *child1;
//
// Construct a new leaf node for character c
//
public:
    node( unsigned char c = 0, int i = -1 ) {
        value = c;
        weight = i;
        child0 = 0;
        child1 = 0;
    }
//
// Construct a new internal node that has
// children c1 and c2.
//
    node( const node* c0, const node *c1 ) {
        value = 0;
        weight = c0->weight + c1->weight;
        child0 = c0;
        child1 = c1;
    }

//
// The comparison operators used to order the
// priority queue.
//
    bool operator>( const node &a ) const {
        return weight > a.weight;
    }

    void traverse( string code = "" )  const;
    void maketable( unsigned code, unsigned bits, TCodeEntry *codetable ) const;
};

//
// The traverse member function is used to
// print out the code for a given node.  It
// is designed to print the entire tree if
// called for the root node.
//

void node::traverse( string code ) const
{
    if ( child0 ) {
        child0->traverse( code + "0" );
        child1->traverse( code + "1" );
    } else {
        cout << " " << (char)((int)value-128) << "      ";
        cout << setw( 2 ) << weight;
        cout << "     " << code << endl;
    }
}

void node::maketable(unsigned code, unsigned bits, TCodeEntry *codetable) const
{
    if ( child0 ) {
        child0->maketable( SetBit(code,bits,0), bits+1, codetable );
        child1->maketable( SetBit(code,bits,1), bits+1, codetable );
    } 
    else {
        codetable[value].code = code;    
        codetable[value].bits = bits;
    }
}

//
// This routine does a quick count of all the
// characters in the input text.  
//

void count_chars(char *text, unsigned n, TCodeEntry *counts )
{
    unsigned i;
    for (i = 0 ; i < 256 ; i++ )
        counts[ i ].count = 0;
    for (i=0; i<n; i++)
        counts[(int)text[i]+128].count++; 
}

TCodeEntry *makecodetable(char *text, unsigned n)
{
    TCodeEntry *result = new TCodeEntry[ 256 ];
    
    count_chars( text, n, result );
/*    unsigned count=0;
    for (int i=0;i<256; i++) {
       count += result[i].count;
       if (result[i].count>0)
          printf("%c: %d\n",i-128,result[i].count);
    }
    printf("count=%d\n",count);*/
    priority_queue< node, vector< node >, greater<node> > q;
//
// First I push all the leaf nodes into the queue
//
    for ( unsigned int i = 0 ; i < 256 ; i++ )
        if ( result[ i ].count )
            q.push(node( i, result[ i ].count ) );
//
// This loop removes the two smallest nodes from the
// queue.  It creates a new internal node that has
// those two nodes as children. The new internal node
// is then inserted into the priority queue.  When there
// is only one node in the priority queue, the tree
// is complete.
//

    while ( q.size() > 1 ) {
        node *child0 = new node( q.top() );
        q.pop();
        node *child1 = new node( q.top() );
        q.pop();
        q.push( node( child0, child1 ) );
    }
//
// Now I compute and return the codetable
//
    //cout << "Char  Symbol   Code" << endl;
    //q.top().traverse();
    q.top().maketable(0u,0u, result);
    return result;
}
